/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
*
* @FILE     li_opdis.c

*
* @ABSTRACT
* This file contains:
*
* @INCLUDE
*/


#include <stdio.h>
#include <string.h>

#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upext.h"

#include "osx_upcon.h"
#include "osx_upext.h"

#include "csi_upcon.h"
#include "csi_uptyp.h"
#include "csi_usext.h"

#include "prs_upcon.h"
#include "prs_upext.h"

#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "gcn_upmes.h"
#include "li_prmes.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"

#include "li_prcon.h"
#include "li_prtyp.h"
#include "li_prext.h"
#include "li_upext.h"

#include "vas_upcon.h"
#include "vas_upext.h"
#include "gcn_upmac.h"


/*
* @LOCALFUN li_ComputeMsgAttributes
*
* @DESCRIPTION Computes message network information
*
* @ABSTRACT
* This function is a subroutine of the Local Intelligence Engine. It is called
* by <@LOCALFUN li_OperationDispatcher> (see previous Operation). The aim of 
* this routine is to compute message network information like direction
* of the CSIDD (leaving, entering CCNDOMAIN) or the kind of exchange mode
* like synchronous response, asynchronous request.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal case) or CSIFAILED (an error
* occurred).
* 
* INOUT   LI_MSGCONTEXT *li_MsgContext_P
*         A pointer to the current message context. The Operation
*         Dispatcher uses and modifies the fields of this context.
*
* OUT     CSIULONG *MsgType
*         Points to the class of transfert for the message.
*
* OUT     CSIULONG *MsgDirection
*         Points to the way the message is going (enter, leave or local).
*
* OUT     CSILONG *returnCode_P
*         Points to the returnvalue (identical).
*
* OUT     CSILONG *reasonCode_P
*         Points to the reason code (error code).
*/
static CSILONG li_ComputeMsgAttributes(
INOUT  LI_MSGCONTEXT   *li_MsgContext_P,
OUT    CSIULONG         *MsgType,
OUT    CSIULONG         *MsgDirection,
OUT    CSILONG          *returnCodeP,
OUT    CSILONG          *reasonCodeP) {

  CSIULONG localMsgDirection = VAS_NOCCN,
           localMsgType      = VAS_NO_MSG;

  CSILONG  ret;
  CSILONG  TargetIsLocal = 0;
    
  TRACE_FUNCTION_ENTRY("li_ComputeMsgAttributes");

  if ( (NULL == returnCodeP) || (NULL == reasonCodeP)) {
    TRACE_FUNCTION_EXIT("li_ComputeMsgAttributes",CSIFAILED,LI_ERR_INTERNAL);
    return CSIFAILED;
  }
 
  ret = CDX_isLocalGateway(li_MsgContext_P->DestGateway,
                           &TargetIsLocal,
                           returnCodeP,
                           reasonCodeP);

  if (CSIFAILED != ret) {
    if (TargetIsLocal) {
      /* Call is local to the Gateway. */
      if (li_MsgContext_P->TraceInfo_P->NbOfPrevGateways == 1) {
        /* Message is local to the gateway if there is only one element */
        /* in the PrevGatewayList. */
        localMsgDirection = VAS_CCNENTER |
                            VAS_CCNEXIT;
      } else {
        if (li_MsgContext_P->Direction == LI_REQUEST) 
          /* The messages arrives on the distant gateway */
          localMsgDirection = VAS_CCNEXIT;
        else
          /* This is the response of service on the distant gateway */
          localMsgDirection = VAS_CCNENTER;
      }
    } else {
      /* Call is starting or coming back to / from the originator. */
      if (li_MsgContext_P->Direction == LI_REQUEST)
        /* The message is starting */
        localMsgDirection = VAS_CCNENTER;
      else
        /* The message is coming back */
        localMsgDirection = VAS_CCNEXIT;
    }
  }

  /* Compute the Message type. Can be one of the following: */
  /* VAS_NO_MSG    : Faulty message, should not happen.     */
  /* VAS_SYNC_REQ  : Synchronous request message            */
  /* VAS_SYNC_RSP  : Synchronous response message           */
  /* VAS_ASYNC_MSG : Asynchronous message                   */
  switch (li_MsgContext_P->ExchangeMode) {
  case CSI_ASYNCHRON:
    localMsgType = VAS_ASYNC_MSG;
    break;
  case CSI_SYNCHRON:
      if (li_MsgContext_P->Direction == LI_REQUEST ) {
        localMsgType = VAS_SYNC_REQ;
      } else
        localMsgType = VAS_SYNC_RSP;
  }

  /* Eval the direction of the traffic, could be enter, leaver, both or */
  /* none of the two options */

  if (MsgType)
    *MsgType = localMsgType;

  if (MsgDirection)
    *MsgDirection = localMsgDirection;

  TRACE_FUNCTION_EXIT("li_ComputeMsgAttributes",*returnCodeP,*reasonCodeP);
  return *returnCodeP;
}


/*
* @LOCALFUN li_OperationDispatcher
*
* @DESCRIPTION Performs a sequence of tasks on the message
*
* @ABSTRACT
* This function is a subroutine of the Local Intelligence Engine. It is
* called either by the Synchronous Processing (<@LOCALFUN li_SyncProc>),
* the Asynchronous Processing (<@LOCALFUN li_AsyncProc>) or the Conversational
* Processing (<@LOCALFUN li_ConvProc>).
*
* The dispatcher performs:
* - the routing by calling <@LOCALFUN li_MsgRoutingScheduler>;
* - the access control by calling <@LOCALFUN li_ACMScheduler>;
* - presentation, data conversion with functions
*  <@LOCALFUN li_GetConvInfo>, <@LOCALFUN li_MsgConvert>;
* - VAS management
*
* Note that the same structure is used as input AND output, so you must 
* have a copy of the message data and associated information (trace info,
* presentation info, quality of service) if you want to restore the initial
* message in case an error occurs.
*
* The execution of each step is conditionned by a flag in the "piloting
* flags"; these flags enable the program to exclude dynamically a task,
* if it appeared that this particular step was not necessary. It is initially
* the same as the required flags.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal case) or CSIFAILED (an error
* occurred).
* 
* @PARAMETER 
* IN     LI_PROCCONTEXT *li_ProcContext_P
*        A pointer to the Local Intelligence Process Context.
*
* IN     CSILONG flags
*        required flags. The following flags can be combined (by | or +):
*        - LI_ROUTING / LI_NO_ROUTING: do or do not perform Message Routing.
*        - LI_ACM / LI_NO_ACM: do or do not perform Access Control.
*        - LI_MSGCONV / LI_NO_MSGCONV: do or do not perform Message Conversion.
*        - LI_VAS / LI_NO_VAS: do or do not perform VAS or VAS script.
*        You may use the default value LI_DEFAULT_FLAGS which is (LI_ROUTING |
*        LI_ACM | LI_MSGCONV | LI_VAS).
*
* INOUT  LI_MSGCONTEXT *li_MsgContext_P
*        A pointer to the current message context. The Operation
*        Dispatcher uses and modifies the fields of this context.
*
* INOUT  CSIPDU *dataDesc_P
*        A pointer to the current message data descriptor. The 
*        Operation Dispatcher uses and modifies its fields.
*
* OUT    CSILONG *returnCode_P
*        Points to the returnvalue (identical).
*
* OUT    CSILONG *reasonCode_P
*        Points to the reason code (error code).
*
* @SEEALSO
* <@LOCALFUN li_SyncProc>
* <@LOCALFUN li_AsyncProc>
*/
/****************************/
/*** OPERATION DISPATCHER ***/
/****************************/
CSILONG   li_OperationDispatcher (
IN     LI_PROCCONTEXT *li_ProcContext_P,
IN     CSILONG         flags,
INOUT  LI_MSGCONTEXT  *li_MsgContext_P,
INOUT  CSIPDU         **dataDesc_PP,
OUT    CSILONG        *returnCode_P,
OUT    CSILONG        *reasonCode_P
)
{
   /*** LOCAL VARIABLES DECLARATIONS ***/
   /************************************/

   CSILONG     pilotFlags = LI_DEFAULT_FLAGS;


   /** Message Conversion variables **/

   CSICHAR32   msgID          = ""; /* Message ID (pres. point of view) */
   CSILONG     codePage       = 0L;   /* Message CodePage                 */
   CSICHAR128  msgHFMT        = ""; /* Message Host Format              */

   CSILONG     outCodePage    = 0L;   /* output CodePage (from op.disp.) */
   CSICHAR128  outHFMT        = ""; /* output HFMT (from op.dispatch.) */

   CSICHAR    *msgXFMT_P      = NULL; /* Message internal syntax */

   CSILONG     dataConvFlag   = FALSE;

   /** Miscellaneous variables **/

   CSILONG     ret    = CSISUCCESS;
   CSILONG     status = LI_ERR_NONE;

   TRACE_FUNCTION_ENTRY("li_OperationDispatcher");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL == returnCode_P) ||
        (NULL==reasonCode_P) ) {
     TRACE_FUNCTION_EXIT("li_OperationDispatcher",CSIFAILED,LI_ERR_INTERNAL);
     return(CSIFAILED);
   }

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /** Copy the required flags in the "piloting flags" **/

   pilotFlags = flags;

   /** Check if VAS are applied... **/

   if ( (0 == (pilotFlags & LI_NO_VAS)) &&
        (0 != strcmp(li_MsgContext_P->QoS_P->VASScript,"")) ) {
     if (0 != (li_MsgContext_P->QoS_P->AppliedQoS & CSI_QOS_VAS_SCRIPT_DONE)) {
       pilotFlags |= LI_NO_VAS;
     }
   } else {
     pilotFlags |= LI_NO_VAS;
   }

   /*** OPERATION DISPATCHING ! ***/
   /*******************************/

   /** STEP 1: MESSAGE ROUTING **/
   /*****************************/

   if ( 0L == ( pilotFlags & LI_NO_ROUTING ) ) {
     if (LI_ERR_NONE == status) {
       ret = li_MsgRoutScheduler(li_ProcContext_P,li_MsgContext_P,*dataDesc_PP,
                                 returnCode_P, reasonCode_P);
       if (CSIFAILED == ret)
         status = *reasonCode_P;
     }
   }

   /** STEP 2: ACCESS CONTROL MANAGEMENT **/
   /***************************************/

   if ( 0L == (pilotFlags & LI_NO_ACM )) {
     if (LI_ERR_NONE == status) {
       ret = li_ACMScheduler(li_ProcContext_P, li_MsgContext_P,
                             returnCode_P,reasonCode_P);

       /* status must contain the first error occured */
       if (CSIFAILED==ret)
         if(LI_ERR_NONE == status) status = *reasonCode_P;
     }
   }

   /** Get type of message **/
/* PBO OR 3945 BEGIN */
/* gettype is always performed in order to process correct stats */

   /*
   if (LI_ERR_NONE == status) 
   { 
   */
    /** BEGIN: MC 1 **/
    if ( ((*dataDesc_PP)->HeaderLen != 0L) ||
         ((*dataDesc_PP)->DataLength != 0L))
    {
       ret = PRES_gettype((CSIDD *)*dataDesc_PP,msgID,&codePage,msgHFMT,
                          returnCode_P, reasonCode_P);
       if ( CSIFAILED== ret ) 
       {
          EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_PRES_ERROR,
                                          "PRES_gettype",(*reasonCode_P)));

          /* status must contain the first error occured */
          if (LI_ERR_NONE == status)
             status = LI_ERR_PRES_GETTYPE;
       }
       else
       {
          strcpy(li_MsgContext_P->MsgTypeId, msgID);
       }
    }
   /*
    }
   */
   /** ENDOF: MC 1 **/

   /** STEP 3: VAS EXECUTION **/
   /***************************************/
   if ( 0L == (pilotFlags & LI_NO_VAS) && (LI_ERR_NONE == status )) {
     CSIULONG MsgType, MsgDirection;

     /* Compute the Vas Msg Attributes */
     ret = li_ComputeMsgAttributes(li_MsgContext_P,
                                   &MsgType, &MsgDirection,
                                   returnCode_P, reasonCode_P);

     if (CSIFAILED != ret) {
       /* Retrieve the msgXFMT for DIRECTORY information */
       /* Check if conversion has to be done or not. */
       if (LI_ERR_NONE == status) {
         /** Allocation of a CSICHAR buffer for the internal syntax string **/

         ret = OS_mem_alloc(sizeof(CSICHAR256), (CSIVOID**)&msgXFMT_P,
                            returnCode_P, reasonCode_P);

         if (CSIFAILED==ret) {
           EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
                                           "OS_mem_alloc",(*reasonCode_P)));

         /* status must contain the first error occured */
         if (LI_ERR_NONE == status)
            status = LI_ERR_ALLOC;

         } else {
           ret = li_GetMsgInfoForVAS(li_MsgContext_P,
                                     msgID,
                                     codePage, msgHFMT,
                                     GATEWAY_CODEPAGE, GATEWAY_HOSTFORMAT,
                                     *dataDesc_PP,
                                     &msgXFMT_P, &dataConvFlag,
                                     returnCode_P,
                                     reasonCode_P);

           /* If no data, it's not necessary to convert the data. */
           if ((TRUE == dataConvFlag ) && ((*dataDesc_PP)->HeaderLen != 0L))
           {
             /* Message should be converted to gateway native format */
             /* As we don't know, we suppose no previous conversion occured */
             /* and data descriptor hold by dataDesc_PP before conversion */
             /* won't be released during conversion */ 
             CSILONG conversion = FALSE;

             li_MsgConvertToPivot(msgXFMT_P, msgID, msgHFMT, codePage,
                &conversion, dataDesc_PP, returnCode_P, reasonCode_P);
             if (*returnCode_P == CSISUCCESS)
             {
               li_MsgConvertFromPivot(msgXFMT_P, msgID, GATEWAY_HOSTFORMAT,
                  GATEWAY_CODEPAGE, &conversion, dataDesc_PP,
                  returnCode_P, reasonCode_P);
             } 

             /* status must contain the first error occured */
             if ((CSIFAILED==*returnCode_P ) && (LI_ERR_NONE == status))
                   status = *reasonCode_P;
           }
         }
       }
     }

     if (VAS_ERR_NONE == status) {
       if ((MsgType != VAS_NO_MSG) && (MsgDirection != VAS_NOCCN)) {
         ret = vas_Execute(li_MsgContext_P->QoS_P,
                           MsgDirection, MsgType,
                           li_MsgContext_P->MsgDesc_P,
                           li_MsgContext_P->EventType,
                           li_MsgContext_P->TraceInfo_P,
                           li_MsgContext_P->TriggerQName,
                           li_MsgContext_P->UserProfileList_P,
                           li_MsgContext_P->PrevGatewayList_P,
                           dataDesc_PP,
                           returnCode_P, reasonCode_P);

         /* status must contain the first error occured */
         if (( CSIFAILED == ret) && (LI_ERR_NONE == status))
           status = *reasonCode_P;
       }
     } else
     {
        *reasonCode_P = status;
     }

     if (LI_ERR_NONE != status) *returnCode_P = CSIFAILED;
     else *returnCode_P = CSISUCCESS;

/*
     if (status == LI_ERR_NONE)
     {
*/
       /* Format may have changed during VAS execution, so get the type */
       /* of the message again */ 
       ret = PRES_gettype((CSIDD *)*dataDesc_PP,msgID,&codePage,msgHFMT,
                            returnCode_P, reasonCode_P);
       if ( CSIFAILED== ret ) 
       {
           EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_PRES_ERROR,
                        "PRES_gettype",(*reasonCode_P)));
          /* status must contain the first error occured */
          if (LI_ERR_NONE == status)
            status = LI_ERR_PRES_GETTYPE;
       }
       else
       {
          strcpy(li_MsgContext_P->MsgTypeId, msgID);
       }
/*
     }
*/
   }
       
   /** STEP 3: MESSAGE CONVERSION **/
   /********************************/

   if ( ((pilotFlags & LI_NO_MSGCONV)==0) &&
        (LI_ERR_NONE == status) && ((*dataDesc_PP)->HeaderLen != 0L) )
   
   { /*** BEGIN : MESSAGE CONVERSION (1) ***/

      /** Get Conversion Information of message **/
      if (msgXFMT_P == NULL)
      {
         ret = OS_mem_alloc(sizeof(CSICHAR256), (CSIVOID**)&msgXFMT_P,
               returnCode_P, reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
            "OS_mem_alloc",(*reasonCode_P)));

            /* status must contain the first error occured */
            if (LI_ERR_NONE == status)
               status = LI_ERR_ALLOC;
         }
      }   

      if (LI_ERR_NONE == status)
      { /** BEGIN: MC 2 **/
      ret = li_GetConvInfo(li_ProcContext_P,li_MsgContext_P,msgID,codePage,
            msgHFMT,*dataDesc_PP,&outCodePage,outHFMT,&msgXFMT_P,&dataConvFlag,
            returnCode_P,reasonCode_P);

      /* status must contain the first error occured */
      if ((CSIFAILED==ret) && (LI_ERR_NONE == status))
      {
         status = *reasonCode_P;
      }

      } /** ENDOF: MC 2 **/

      /** Conversion if needed **/

      if (LI_ERR_NONE == status)
      { /** BEGIN: MC 3 **/
      
      if (TRUE==dataConvFlag)
      {
         ret = li_MsgConvert(li_MsgContext_P, msgXFMT_P, msgID, msgHFMT, 
	       codePage, outHFMT, outCodePage, dataDesc_PP,
               returnCode_P,reasonCode_P);

         /* status must contain the first error occured */
         if ((CSIFAILED==ret) && (LI_ERR_NONE == status))
         {
            status = *reasonCode_P;
         }
      }

      } /** ENDOF: MC 3 **/

   } /*** ENDOF : MESSAGE CONVERSION (1) ***/

   /* OR 3062 - Fix memory leaks */

   if (NULL != msgXFMT_P)
   {
      OS_mem_free((CSIVOID**)&msgXFMT_P, &ret, reasonCode_P);
   }

   /* OR 3062 */
   /* OR 4654_1 */
   /* following the TA suppression, the LI has to set the internal flag  */
   /* indicating if the file has to be sent by memory or not             */

   if ((0 != (li_MsgContext_P->XferMemory & CSI_DD_FLAGS_XFER_BY_MEMORY)) &&
       (CSI_DD_FLAGS_FILE_STORAGE(*dataDesc_PP)))
   {
       /* Request/Response and Asynchronous case */
       /* if the message is a request (to   remote gateway)  OR    */
       /*                   a reply   (from remote gateway)        */
          if (   (   (    (CSI_REMOTE_EVENT        == li_MsgContext_P->EventType)
                       || (CSI_DIRECT_REMOTE_EVENT == li_MsgContext_P->EventType)
                     )
                  && (LI_REQUEST      == li_MsgContext_P->Direction)     
                 )
              || (   (CSI_LOCAL_EVENT == li_MsgContext_P->EventType)
                  && (LI_REPLY        == li_MsgContext_P->Direction)
                 )
             )
          {
              /* Set the XFER flag to XFER_BY_MEMORY in order to force
                 the data to be sent by memory if they are located into files.  */
               CSI_DD_FLAGS_SET_XFER_BY_MEMORY(*dataDesc_PP);
          }
   }


   *reasonCode_P = status;

   if(LI_ERR_NONE == status)
      *returnCode_P = CSISUCCESS;
   else
      *returnCode_P = CSIFAILED;

   
   TRACE_FUNCTION_EXIT("li_OperationDispatcher",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}
 
/*****************************************************************************/
