/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   CCN
* @PACKAGE  LI (Local Intelligence engine)
*
* @FILE     li_synch.c

* @ABSTRACT
* The Local Intelligence SYNCHronous processing routines.
*
* This file contains:
*
* @INCLUDE
*/


#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upext.h"

#include "rou_upcon.h"
#include "rou_upext.h"

#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"
#include "cdx_upmac.h"

#include "csi_upcon.h"
#include "csi_uptyp.h"
#include "csi_usext.h"

#include "spi_upcon.h"

#include "qos_upmac.h"

#include "loglo_uptyp.h"
#include "evr_upext.h"
#include "evr_upmac.h"
#include "evr_usmes.h"
#include "gcn_upmes.h"
#include "li_prmes.h"
#include "gcn_upmac.h"

#include "li_prcon.h"
#include "li_prtyp.h"
#include "li_prext.h"
#include "li_upext.h"

/*************************************************************/
/*** THE LOCAL INTELLIGENCE SYNCHRONOUS PROCESSING ROUTINE ***/
/*************************************************************/
/*
* @LOCALFUN li_SyncProc
*
* @DESCRIPTION Processes the synchronous exchanges
*
* @ABSTRACT
* This is the Synchronous Processing routine of the Local Intelligence
* engine. It is called by the li_service routine upon reception of a
* synchronous message. This routine coordinates the different tasks (by
* calling the Operation Dispatcher), then calls the next local destination
* (either RAP/GAS or a LI), then processes the reply.
*
* What it does:
* - call <@LOCALFUN li_OperationDispatcher>, to perform message routing,
*   access control, VAS management and data presentation;
* - rebuild the trace information <@FUNCTION gcn_BuildTraceInfo>;
* - call the next entity with <@LOCALFUN li_call>;
* - call <@LOCALFUN li_OperationDispatcher>, this time for the reply.
*
* @RETURNVALUE CSILONG
* Possible returnvalues are: CSISUCCESS (normal) or CSIFAILED (error).
*
* @PARAMETER
* IN     LI_PROCCONTEXT *li_ProcContext_P
*        The Local Intelligence process context.
*
* INOUT  CSIPDU **request_PP
*        The request Data Descriptor.
*
* INOUT  LI_MSGCONTEXT *li_MsgContext_P
*        The currently processed message context.
*
* OUT    CSIPDU reply_PP
*        The reply Data Descriptor.
*
* OUT    CSILONG *returnCode_P
*        Points to either CSISUCCESS (normal) or CSIFAILED (error).
*
* OUT    CSILONG *reasonCode_P
*        In case of error, indicates the associated error code.
*
* OUT    CSILONG *svcRetCode_P
*        A specific Service Return Code returned by the distant service.
*/
CSILONG  li_SyncProc(
IN       LI_PROCCONTEXT  *li_ProcContext_P,
INOUT    CSIPDU          **request_PP,
INOUT    LI_MSGCONTEXT   *li_MsgContext_P,
OUT      CSIPDU          **reply_PP,
OUT      CSILONG         *returnCode_P,
OUT      CSILONG         *reasonCode_P,
OUT      CSILONG         *svcRetCode_P
)
{
   /*** LOCAL VARIABLES DECLARATION ***/
   /***********************************/

   /** Miscellaneous local variables **/

   CSILONG   ret                  = CSISUCCESS;
   CSILONG   status               = LI_ERR_NONE;
   CSILONG   tmpRet               = CSISUCCESS;
   CSILONG   tmpReas              = LI_ERR_NONE;

   CSIPDU   *orgReply_P           = NULL;


   CSIBOOLEAN isOrigLI            = CSIFALSE; /* for audit purposes */
   CSIBOOLEAN firstLISkipped      = CSIFALSE; /* for http2csi       */

   /* OR7273 : Original message length, as sent by AP. */
   /* Used if data presentation takes place.           */
   CSIULONG originalMsgLength;

   CSICHAR8 trctoGwName;
   CSILONG  trcReturnCode = CSISUCCESS;
   CSILONG  trcReasonCode = LI_ERR_NONE;

   struct timeval tv1;
   struct timezone tz1;

   struct timeval tv2;
   struct timezone tz2;

   CSILONG  ccnLocalVersion = CSIIITFHDR_VERSION_1;
   CSILONG  ccnPartnerVersion = CSIIITFHDR_VERSION_1;
   CSILONG t4, t4U, t3, t3U, t2, t2U, t1, t1U;
   CSILONG  ccnEventType;
   CSICHAR48  tt1;
   CSICHAR48  tt2;
   CSICHAR48  tt3;
   CSICHAR48  tt4;
   CSICHAR256 szServiceDN = {""};
   struct tm *psTime;
   struct timeval tim;

   float responseTime;

   TRACE_FUNCTION_ENTRY("li_SyncProc");

   /*** PARAMETERS VERIFICATION ***/
   /*******************************/

   if ( (NULL==returnCode_P)||(NULL==reasonCode_P) )
   {
      TRACE_FUNCTION_EXIT("li_SyncProc",CSIFAILED,LI_ERR_INTERNAL);
      return(CSIFAILED);
   }

   /*** INITIALIZATIONS ***/
   /***********************/

   *returnCode_P = CSISUCCESS;
   *reasonCode_P = LI_ERR_NONE;

   /* As EventType may change, we must keep in isOrigLI that this LI */
   /* is the first one (or not).                                     */
   isOrigLI = (li_MsgContext_P->EventType == CSI_LOCAL_EVENT)
              ? CSITRUE
              : CSIFALSE;
  /* This is true when the call the the LI is made by http2csi gas
   * In that case we don't go to the first LI but we directly go to
   * the second LI
   */
   firstLISkipped = (li_MsgContext_P->EventType == CSI_DIRECT_REMOTE_EVENT)
                    ? CSITRUE
                    : CSIFALSE;

   originalMsgLength = (*request_PP)->DataLength;

   t1 = t1U = t2 = t2U = t3 = t3U = t4 = t4U = 0L;

   responseTime = 0.00;

   /*** RETRIEVING THE RESOURCE RUNMODE FOR AUDIT  ***/
   /**************************************************/

   switch (((CSIDD *)*request_PP)->Flags & CSI_DD_FLAGS_MODE_SET)
   {
      case CSI_DD_FLAGS_PRODUCTION_MSG:
         strcpy(li_MsgContext_P->RunMode,CDX_C_PRODUCTION_STR);
         break;
      case CSI_DD_FLAGS_BACKUP_MSG:
         strcpy(li_MsgContext_P->RunMode,CDX_C_BACKUP_STR);
         break;
      case CSI_DD_FLAGS_TEST_MSG:
         strcpy(li_MsgContext_P->RunMode,CDX_C_TEST_STR);
         break;
      case CSI_DD_FLAGS_MAINTENANCE_MSG:
         strcpy(li_MsgContext_P->RunMode,CDX_C_MAINTENANCE_STR);
         break;
   }


   /*** CALLING THE OPERATION DISPATCHER FOR REQUEST ***/
   /****************************************************/
   /** All tasks are required: routing,access control,conversions, **/
   /** VAS executions.                                             **/
                                                      
   if (LI_ERR_NONE == status)                         
   {                        
      /* Ensure the Direction indicates REQUEST data and HFMT is NULL    */
      /* This is necessary in case of a degraded msg "switch back"       */
      /* because the C.P. and HFMT of degraded queue are not necessarily */
      /* those of the initial client application...                      */
      li_MsgContext_P->Direction           = LI_REQUEST;
      li_MsgContext_P->ClientHostFormat[0] = '\0';    
      li_MsgContext_P->ServerHostFormat[0] = '\0';    
                                                 
      ret = li_OperationDispatcher(li_ProcContext_P, LI_DEFAULT_FLAGS,
            li_MsgContext_P, request_PP, returnCode_P, &tmpReas);
                                                      
      if (CSIFAILED==ret)                             
      {                 
         *reasonCode_P = tmpReas;
         status = *reasonCode_P;
      }                         
   }

   /*** GET PROTOCOL VERSION OF LOCAL GATEWAY ***/
   /*********************************************/

   if (LI_ERR_NONE == status)
   {
      if (CSISUCCESS != CDX_getCCNProtocolVersion(NULL
                                                 ,&ccnLocalVersion
                                                 ,returnCode_P
                                                 ,reasonCode_P
                                                 )
         )
      {
         status = *reasonCode_P;
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
                         "CDX_getCCNProtocolVersion",(*reasonCode_P)));
      }
      else
      {
         /* check if protocol version is at least 3 */
         if (ccnLocalVersion <= LI_PROTOCOL_VERSION_2)
         {
            status = LI_ERR_INTERNAL;
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
                         "CDX_getCCNProtocolVersion",(status)));
         }
      }
   }

   /*** GET PROTOCOL VERSION OF PARTNER GATEWAY ***/
   /***********************************************/

   if (LI_ERR_NONE == status)
   {
      if (CSISUCCESS != CDX_getCCNProtocolVersion((isOrigLI == CSITRUE)
                                                       ?li_MsgContext_P->DestGateway
                                                       :li_MsgContext_P->OrigGateway
                                                 ,&ccnPartnerVersion
                                                 ,returnCode_P
                                                 ,reasonCode_P
                                                 )
         )
      {
         status = *reasonCode_P;
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_CDX_ERROR,
                      "CDX_getCCNProtocolVersion",(*reasonCode_P)));
      }
      else
      {
         /* if the Destination is at Version 3 or higher we can send the new traceinfo */
         if (ccnPartnerVersion > LI_PROTOCOL_VERSION_2)
               li_MsgContext_P->TraceInfo_P->Version = CSIIITFHDR_VERSION;
         else
               li_MsgContext_P->TraceInfo_P->Version = CSIIITFHDR_VERSION_1;
      }
   }

   /*** INPUT AUDIT TRACE : JUST PREPARING ***/
   /******************************************/

   if (isOrigLI == CSITRUE)
   {
      CSICHAR48 toGwName = {""};

      if (li_MsgContext_P->EventType == CSI_REMOTE_EVENT)
         CDX_extractNameFromDN(li_MsgContext_P->DestGateway,toGwName);
        
      strcpy(trctoGwName, toGwName);
      trcReturnCode = ((status==LI_ERR_NONE)? CSISUCCESS : CSIFAILED);
      trcReasonCode = status;
   }

   /**                             **/
   /** Taking the first timestamps **/
   /**                             **/
   if (isOrigLI == CSITRUE)          
   {                                 
      gettimeofday(&tim, NULL);
      li_MsgContext_P->TraceInfo_P->SubmitTimeSec = tim.tv_sec;
      li_MsgContext_P->TraceInfo_P->SubmitTimeUSec = tim.tv_usec;
      t1  = tim.tv_sec;                      
      t1U = tim.tv_usec;                     
      if (LI_ERR_NONE != status)
      {                        
          /* we won't try to contact next LI */
          t4 = t3 = t2 = t1;                 
          t4U = t3U = t2U = t1U;             
      }                        
   }                           
   else /* second LI */
   {                  
      gettimeofday(&tim, NULL);
      t1  = li_MsgContext_P->TraceInfo_P->SubmitTimeSec;
      t1U = li_MsgContext_P->TraceInfo_P->SubmitTimeUSec;
      t2  = tim.tv_sec;                      
      t2U = tim.tv_usec;                     
      li_MsgContext_P->TraceInfo_P->SubmitTimeSec = t2;
      li_MsgContext_P->TraceInfo_P->SubmitTimeUSec = t2U;
      if (LI_ERR_NONE != status)             
      {                                      
          /* we won't try to contact app server */
          t3 = t2;                           
          t3U = t2U;                         
      }            
   }


   /*** CALLING THE NEXT ENTITY ***/
   /*******************************/

   /**                                **/
   /** Building the Trace Information **/
   /**                                **/
   if (LI_ERR_NONE == status)
   {
      ret = gcn_BuildTraceInfo(li_MsgContext_P->EventType, 
            li_MsgContext_P->TraceInfo_P,
            NULL,li_MsgContext_P->UserProfileList_P,
            li_MsgContext_P->PrevGatewayList_P,*request_PP,
            returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
         "gcn_BuildTraceInfo",(*reasonCode_P)));
         status        = LI_ERR_INTERNAL;
      }
   }

   /**                                      **/
   /** Allocating the Reply Data Descriptor **/
   /**                                      **/
   if (LI_ERR_NONE == status)
   {
      ret = gcn_AllocDataDesc(0L,reply_PP,returnCode_P,reasonCode_P);
      if (CSIFAILED==ret)
      {
         EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_PRES_ERROR,
                      "gcn_AllocDataDesc(reply_P)",(*reasonCode_P)));
         status        = LI_ERR_ALLOC;
      }
      else
      {
        /* Keep the DD pointer in order to release it after the 
           operation dispatcher call. */
        orgReply_P = *reply_PP;
      }
   }

   /**                                                                 **/ 
   /** CSI_call'ing (customized CSI_call, see li_call) the next entity **/
   /**                                                                 **/ 
   if (LI_ERR_NONE == status)
   {
         /* Going... going... gone ! */

         gettimeofday(&tv1, &tz1);
         ret = li_call(li_MsgContext_P->NextHopName,(CSIDD *)*request_PP,
               (CSITPNOBLOCK + CSITPSIGRSTRT),
               li_MsgContext_P->QoS_P,(CSIDD *)*reply_PP,
               svcRetCode_P, returnCode_P, reasonCode_P);
         gettimeofday(&tv2, &tz2);

         /* Now calculate the response time */
         responseTime = tv2.tv_sec - tv1.tv_sec;
         responseTime += ((float)(tv2.tv_usec - tv1.tv_usec)) / (float)1000000.0000;
         /* OR5139 : prevent LI to insert negative response times */
         /* when GWs' clocks are synchronised                     */
         responseTime = (responseTime >= 0.00 ? responseTime : 0.50);


         /**                                                                    **/
         /** Taking the last timestamps and retreive other ones from trace info **/
         /**                                                                    **/
         if( isOrigLI == CSITRUE )
         {
            gettimeofday(&tim, NULL);
            t4 = tim.tv_sec;
            t4U = tim.tv_usec;

            /* if we received this from a remote LI */
            if( li_MsgContext_P->TraceInfo_P->NbOfPrevGateways > 1 )
            {
               /* default value in case li_call failed */
               /* no app server time: t2 = t3 and all the execution time is considered backbone time */
               t2 = t3 = t4;
               t2U = t3U = t4U;

               /* check if the li_call succeeded */
               if (CSISUCCESS == ret)
               {
                  /* default value in case of problem to retreive t2 and t3 from trace info */
                  t2 = t3 = t2U = t3U = 0L;

                 /* We can only retrieve the trace header if our destination
                  * is set at ccnProtocolVersion higher than 2 */
                  if( ccnPartnerVersion > LI_PROTOCOL_VERSION_2 )
                  {
                     if (CSISUCCESS !=  gcn_GetTraceInfo(GCN_GET,*reply_PP,&ccnEventType
                                                        ,li_MsgContext_P->TraceInfo_P
                                                        ,li_MsgContext_P->TriggerQName
                                                        ,NULL
                                                        ,NULL
                                                        ,returnCode_P
                                                        ,reasonCode_P
                                                        )
                        )
                     {
                        /* could not retreive trace info */
                        EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_PRES_ERROR,
                                     "gcn_GetTraceInfo",(*reasonCode_P)));
                        status = *reasonCode_P;
                     }
                     else
                     {
                        t2  = li_MsgContext_P->TraceInfo_P->SubmitTimeSec;
                        t2U = li_MsgContext_P->TraceInfo_P->SubmitTimeUSec;
                        t3  = li_MsgContext_P->TraceInfo_P->RecvTimeSec;
                        t3U = li_MsgContext_P->TraceInfo_P->RecvTimeUSec;
                     }
                  }
               }
            }
            else /* loopback */
            {
               /* no backbone time: t2 = t1 and t3 = t4 */
               t2 = t1;
               t2U = t1U;
               t3 = t4;
               t3U = t4U;
            }
         }
         else
         {
            /* We are on the second LI, generate t3 and insert it into the traceinfo */
            gettimeofday(&tim, NULL);
            t3U = tim.tv_usec;
            t3 = tim.tv_sec;
            li_MsgContext_P->TraceInfo_P->SubmitTimeSec = t2;
            li_MsgContext_P->TraceInfo_P->SubmitTimeUSec = t2U;
            li_MsgContext_P->TraceInfo_P->RecvTimeSec = t3;
            li_MsgContext_P->TraceInfo_P->RecvTimeUSec = t3U;
            if (firstLISkipped)
            {  
              /* take the last timestamp
               * in case we enterded with CSI_DIRECT_REMOTE_EVENT
               * which is the case for http2csi gas
               * we don't go back the first LI in this case
               */
               gettimeofday(&tim, NULL);
               t4 = tim.tv_sec;
               t4U = tim.tv_usec;
            }
         }

         /**                      **/
         /** Ckeck li_call result **/
         /**                      **/
         if (CSIFAILED==ret)
         {
              /** Houston, we've got a problem... **/
               if (TUX_ERR_SVCFAIL != *reasonCode_P)
               { 
                  /* Only trace if error is not a logical SVC failure */

                  EVR_ERRORLOG(EVR_MAJOR,(LI_MSG_CALL_FAILED,"next",
                  li_MsgContext_P->NextHopName,(*reasonCode_P)));
               }

               if(TUX_ERR_NOENT == *reasonCode_P)
                  *reasonCode_P = LI_ERR_SVC_NOT_AVAILABLE;
   
               /* Delete file, since no more needed */
               gcn_DeleteFile(*request_PP, &tmpRet, &tmpReas);
   
               status        = *reasonCode_P;
         }
         else /* li_call failed */
         {
            EVR_TRACELOG(EVR_DEBUG,(LI_TRC_SVC_CALLED,
            li_MsgContext_P->TraceInfo_P->UniqueId,
            li_MsgContext_P->NextHopName));
   
            /* Delete file, since no more needed */
            gcn_DeleteFile(*request_PP, &tmpRet, &tmpReas);
   
         }
   }

   /*** INPUT AUDIT TRACE (REQUEST) ***/
   /***********************************/

   li_PrepAudit(t1, t1U, tt1);
   li_PrepAudit(t2, t2U, tt2);
   li_PrepAudit(t3, t3U, tt3);
   li_PrepAudit(t4, t4U, tt4);

   if (isOrigLI == CSITRUE)
   {

       EVR_SYNC_AUDIT_IN_FROM_AP(t1,
                           (0 == strcmp("", trctoGwName) ? getenv(CDX_C_ENV_GATEWAY) : trctoGwName),
                           li_MsgContext_P->TraceInfo_P->UniqueId,
                           li_MsgContext_P->MsgTypeId,
                           originalMsgLength,
                           trcReturnCode,
                           trcReasonCode,
                           li_MsgContext_P->TraceInfo_P->OrgUserName,
                           li_MsgContext_P->TraceInfo_P->OrgAppliName,
                           li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                           AUDIT_KW_REQ,
                           responseTime,
                           *svcRetCode_P,
			   tt1,
			   tt2,
			   tt3,
			   tt4,
                           CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));
   }

   /*** OUTPUT AUDIT TRACE (REQUEST) ***/
   /************************************/

   if (li_MsgContext_P->EventType == CSI_LOCAL_EVENT)
   {
      CSICHAR8 fromGwName = {""};

      CDX_extractNameFromDN(li_MsgContext_P->OrigGateway,
                            fromGwName);

      EVR_SYNC_AUDIT_OUT_TO_AP(t1,
                         (0 == strcmp("", fromGwName) ? getenv(CDX_C_ENV_GATEWAY) : fromGwName),
                         li_MsgContext_P->TraceInfo_P->UniqueId,
                         li_MsgContext_P->MsgTypeId,
                         (*request_PP)->DataLength,
                         *returnCode_P,
                         *reasonCode_P,
                         li_MsgContext_P->TraceInfo_P->OrgUserName,
                         li_MsgContext_P->TraceInfo_P->OrgAppliName,
                         li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                         AUDIT_KW_REQ,
                         responseTime,
                         *svcRetCode_P,
                         tt1,
                         tt2,
                         tt3,
                         tt4,
                         CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));

    }
    /*--- END OUTPUT AUDIT TRACE ---*/

   /*** CALLING THE OPERATION DISPATCHER FOR REPLY ***/
   /**************************************************/
   /** Neither message routing nor access control (already done) **/

   /* reply_PP can be null in some error cases */
   originalMsgLength = (NULL != reply_PP && NULL != *reply_PP)
                     ? (*reply_PP)->DataLength
                     : 0L;

   if (LI_ERR_NONE == status)
   {
      /* Change "Direction", now working on REPLY data */
      li_MsgContext_P->Direction = LI_REPLY;
      li_MsgContext_P->TraceInfo_P->Version = CSIIITFHDR_VERSION_1;

      if ( isOrigLI == CSIFALSE )
      {
         /* if we our destination is at version 3 we will pass IIIFTH version 2 */
         if( ccnPartnerVersion > LI_PROTOCOL_VERSION_2 )
            li_MsgContext_P->TraceInfo_P->Version = CSIIITFHDR_VERSION;
      }
         
      ret = li_OperationDispatcher(li_ProcContext_P,(LI_NO_ROUTING+LI_NO_ACM),
            li_MsgContext_P, reply_PP, returnCode_P, &tmpReas);

      /* Write the IIFTHEADER into the reply so that the first LI
       * can audit all the four timestamps
       */
      if (    isOrigLI == CSIFALSE
           && ccnPartnerVersion > LI_PROTOCOL_VERSION_2
           && firstLISkipped == CSIFALSE
         )
      {
         ret = gcn_BuildTraceInfo(li_MsgContext_P->EventType,
                                  li_MsgContext_P->TraceInfo_P,
                                  NULL,li_MsgContext_P->UserProfileList_P,
                                  li_MsgContext_P->PrevGatewayList_P,*reply_PP,
                                  returnCode_P,reasonCode_P);
         if (CSIFAILED==ret)
         {
            EVR_ERRORLOG(EVR_UNDETERMINATE,(LI_MSG_GCN_ERROR,
            "gcn_BuildTraceInfo",(*reasonCode_P)));
            status        = LI_ERR_INTERNAL;
         }
      }   

      if (CSIFAILED==ret)
      {
         *reasonCode_P = tmpReas;
         status = *reasonCode_P;
      }

      /* Release the DD previously allocated if the resulting DD is different. */
      if ((orgReply_P != NULL) && (*reply_PP != orgReply_P))
      {
        gcn_FreeDataDesc(&orgReply_P, &tmpRet, &tmpReas); 
      }
   }

   /*** THE END ***/
   /***************/

   if (LI_ERR_NONE != status)
       *returnCode_P = CSIFAILED;
   else
       *returnCode_P = CSISUCCESS;

   *reasonCode_P = status;

   /*** INPUT AUDIT TRACE (REPLY) ***/
   /*********************************/

   li_PrepAudit(t1, t1U, tt1);
   li_PrepAudit(t2, t2U, tt2);
   li_PrepAudit(t3, t3U, tt3);
   li_PrepAudit(t4, t4U, tt4);

   if (li_MsgContext_P->EventType == CSI_LOCAL_EVENT)
   {
      CSICHAR8 fromGwName = {""};

      CDX_extractNameFromDN(li_MsgContext_P->OrigGateway, fromGwName);

      EVR_SYNC_AUDIT_IN_FROM_AP(t1,
                         (0 == strcmp("", fromGwName) ? getenv(CDX_C_ENV_GATEWAY) : fromGwName),
                         li_MsgContext_P->TraceInfo_P->UniqueId,
                         li_MsgContext_P->MsgTypeId,
                         originalMsgLength,
                         *returnCode_P,
                         *reasonCode_P,
                         li_MsgContext_P->TraceInfo_P->OrgUserName,
                         li_MsgContext_P->TraceInfo_P->OrgAppliName,
                         li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                         AUDIT_KW_RSP,
                         responseTime,
                         *svcRetCode_P,
			 tt1,
			 tt2,
			 tt3,
			 tt4,
                         CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));
   }

   /*** OUTPUT AUDIT TRACE ***/
   /**************************/

   if (isOrigLI == CSITRUE)
   {
      CSICHAR8 toGwName = {""};

      if (li_MsgContext_P->EventType == CSI_REMOTE_EVENT)
         CDX_extractNameFromDN(li_MsgContext_P->DestGateway, toGwName);

      EVR_SYNC_AUDIT_OUT_TO_AP(t1,
                         (0 == strcmp("", toGwName) ? getenv(CDX_C_ENV_GATEWAY) : toGwName),
                         li_MsgContext_P->TraceInfo_P->UniqueId,
                         li_MsgContext_P->MsgTypeId,
                         (NULL != reply_PP && NULL != *reply_PP) ? (*reply_PP)->DataLength : 0L,
                         *returnCode_P,
                         *reasonCode_P,
                         li_MsgContext_P->TraceInfo_P->OrgUserName,
                         li_MsgContext_P->TraceInfo_P->OrgAppliName,
                         li_MsgContext_P->TraceInfo_P->TrgtCCNServiceName,
                         AUDIT_KW_RSP,
                         responseTime,
                         *svcRetCode_P,
			 tt1,
			 tt2,
			 tt3,
			 tt4,
                         CDX_GTWMODE_PREFIX_CSICHAR(li_MsgContext_P->RunMode));
   }

   TRACE_FUNCTION_EXIT("li_SyncProc",*returnCode_P,*reasonCode_P);

   return(*returnCode_P);
}

/***********************************/
/*** LI CUSTOMIZED CALL FUNCTION ***/
/***********************************/
/*
* @LOCALFUN li_call
*
* @DESCRIPTION A 'customized' CSI_call for the LI
*
* @ABSTRACT
*
* This is a customized <@FUNCTION CSI_call> (same interface). When calling
* a RAP, if an error SPI_ERR_FATAL occurs, this function tries calling
* again the service with the same parameters. The maximum number of retrys 
* is defined by LI_NB_OF_RETRYS. 
*
* Rem : As the error SPI_ERR_FATAL may be currently returned by a RAP
* ----- only, there is no need to check the destination of the call.
*       However, may this change, it would be necessary to perform
*       additional verifications.
*
* @RETURNVALUE CSILONG
*
* @PARAMETER
* The same as <@FUNCTION CSI_call>.
*
* @SEEALSO
* <@FUNCTION CSI_call>
*/
CSILONG li_call (
IN      CSICHAR48  svcName,
IN      CSIDD     *inDataDesc_P,
IN      CSILONG    flags,
INOUT   CSIQOS    *qos_P,
OUT     CSIDD     *outDataDesc_P,
OUT     CSILONG   *svcRetCode_P,
OUT     CSILONG   *returnCode_P,
OUT     CSILONG   *reasonCode_P
)
{
   /** LOCAL VARIABLES DECLARATION AND INITIALISATION **/
   /****************************************************/

   CSILONG     calls_counter = 0L;
   CSIBOOLEAN  stop_calling  = CSIFALSE;

   /** CALL LOOP **/
   /***************/

   while ( (stop_calling == CSIFALSE) && 
           (calls_counter <= LI_NB_OF_RETRYS) )
   {
      CSI_call(svcName, inDataDesc_P, flags, qos_P, outDataDesc_P,
               svcRetCode_P, returnCode_P, reasonCode_P);

      /* If no error SPI_ERR_FATAL occurred, stop calling */

      if ( *reasonCode_P != SPI_ERR_FATAL ) stop_calling = CSITRUE;

      /* (else) increment counter */

      calls_counter++ ;
   }

   /* Identify internal errors (all errors except TUX_ERR_SVCFAIL) */
   /****************************************************************/
   /* Internal errors are tagged by a specific flag in the QoS */

   if (TUX_ERR_SVCFAIL != *reasonCode_P) QOS_SET_CCN_ERR(qos_P);

   return (*returnCode_P);
}

/*
* @LOCALFUN li_ResolveReplyQ 
*
* @DESCRIPTION Convert a reply 2 queue name into a real target name 
*
* @ABSTRACT
*
* replyQ is parsed and @local @localgw s replaced by the current gateway
* and site name. if a FQDN is wanted, the current CCN gateway name is
* inserted before the siteName, of course, only is siteName is equal to 
* the current site. 
*
* @RETURNVALUE CSIVOID
*
* @PARAMETER
* IN    CSIBOOLEAN returnFQDN	if TRUE, return a FQDN if possible
* INOUT CSICHAR48  replyQ       the queue name to convert 
*
* @SEEALSO
* <@FUNCTION CSI_call>
*/
CSIVOID li_ResolveReplyQ(
  IN    CSIBOOLEAN returnFQDN,
  INOUT CSICHAR48 replyQ)
{
   CSICHAR      *arrobase;
   CSICHAR48    szNameCopy;

   arrobase = strchr (replyQ, ADDRESS_SEPARATOR[0]);
   if (arrobase == NULL)
   {
      if ('\0' == replyQ[0])
         return;

      /* No site name specified */
      strcat(replyQ, ADDRESS_SEPARATOR);

      if (TRUE == returnFQDN)
      {
         strcat(replyQ, getenv(CDX_C_ENV_GATEWAY));
         strcat(replyQ, ".");
      }

      strcat(replyQ, getenv(CDX_C_ENV_SITE));
   }
   else
   {
      strcpy(szNameCopy, replyQ);
      arrobase = strchr (szNameCopy, ADDRESS_SEPARATOR[0]); 
      (*arrobase) = '\0';
      arrobase++; 

      if ('\0' == szNameCopy[0])
      {
         strcpy(replyQ, "");
         return;
      }

      strcpy(replyQ, szNameCopy);
      strcat(replyQ, ADDRESS_SEPARATOR);
     
      if ( (0 == strcasecmp(LOCAL_GTW_ADDRESS, arrobase))  || 
           ( (0 == strcasecmp(LOCAL_SITE_ADDRESS, arrobase)) && (TRUE == returnFQDN)) )
      {
         strcat(replyQ, getenv(CDX_C_ENV_GATEWAY));
         strcat(replyQ, ".");
         strcat(replyQ, getenv(CDX_C_ENV_SITE));
      }
      else if (0 == strcasecmp(LOCAL_SITE_ADDRESS, arrobase))
      {
         strcat(replyQ, getenv(CDX_C_ENV_SITE)); 
      }
      else if (FALSE == returnFQDN)
      {
         /* Only recopy the arrobase */
         strcat(replyQ, arrobase);
      }
      else if (0 == strcasecmp(arrobase, getenv(CDX_C_ENV_SITE)))
      {
         /* if not a FQDN, and arrobase = current site */
         /* add current gateway                        */
         strcat(replyQ, getenv(CDX_C_ENV_GATEWAY));
         strcat(replyQ, ".");
         strcat(replyQ, getenv(CDX_C_ENV_SITE));
      }
      else  
      {
         strcat(replyQ, arrobase);
      }
   }
}

CSIVOID li_PrepAudit(
IN       CSILONG         lTime,
IN       CSILONG         lUTime,
INOUT    CSICHAR48       cTime)
{
   CSICHAR         szAuditSubmitTime[15];
   time_t          sTimeT;
   struct tm       *psTime;

    if( lTime > 0 )
    {
     
       sTimeT = (time_t)lTime;
       psTime = localtime(&sTimeT);

       sprintf(szAuditSubmitTime, "%.4d%.2d%.2d%.2d%.2d%.2d",
              psTime->tm_year + 1900,
              psTime->tm_mon+1,
              psTime->tm_mday,
              psTime->tm_hour,
              psTime->tm_min,
              psTime->tm_sec);
       sprintf(cTime,"%s.%06d",szAuditSubmitTime,lUTime);
     }
     else
     {   
        strcpy(cTime,"nil");
     }
}

/*****************************************************************************/
